from ctypes import c_byte, c_double, c_int, c_ushort
from typing import Tuple

import forcedimension_core.runtime as _runtime
from forcedimension_core.typing import (
    Array, SupportsPtr, SupportsPtrs3, c_double_ptr
)


def getDeltaEncoders(out: SupportsPtrs3[c_int], ID: int = -1) -> int:
    """
    Read all encoders values of the DELTA structure.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_int] out:
        An output buffer to store the delta encoder values.

    :raises IndexError:
        If ``len(out) < 3``.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, and -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc3`
    | :class:`forcedimension_core.containers.numpy.Enc3`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncoderToPosition()`
    """

    return _runtime._libdhd.dhdGetDeltaEncoders(*out.ptrs, ID)


def getWristEncoders(out: SupportsPtrs3[c_int], ID: int = -1) -> int:
    """
    Read all encoders values of the wrist structure.


    Note
    ----
    This feature only applies to devices with a wrist. See
    the :ref:`device_types` section for more details.


    :param SupportsPtrs3[ctypes.c_int] out:
        An output buffer to store the wrist encoder values.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc3`
    | :class:`forcedimension_core.containers.numpy.Enc3`
    | :func:`forcedimension_core.dhd.wristEncodersToOrientation()`
    | :func:`forcedimension_core.dhd.wristEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.wristOrientationToEncoder()`
    """

    return _runtime._libdhd.dhdGetWristEncoders(*out.ptrs, ID)


def deltaEncoderToPosition(
    enc: Array[int, int],
    out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the position of the end-effector about
    the X, Y, and Z axes (in [m]) for a given set of encoder values.

    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to raw encoder values on axis 0, 1, and 2,
        respectively.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the position about the X, Y, and Z axes
        (in [m]).

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaEncoders()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaPositionToEncoder()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncodersToJointAngles()`
    """

    return _runtime._libdhd.dhdDeltaEncoderToPosition(
        enc[0], enc[1], enc[2], *out.ptrs, ID
    )


def deltaPositionToEncoder(
    pos: Array[int, float],
    out: SupportsPtrs3[c_int],
    ID: int = -1,
) -> int:
    """
    Computes and returns the encoder values of the end-effector
    for a given end-effector position.

    :param Array[int, float] pos:
        Sequence of ``(px, py, pz)`` where ``px``, ``py``, and ``pz``
        refer to the end-effector position on the X, Y, and Z axes,
        respectively (in [m]).

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_int] out:
        An output buffer to store the delta encoder values.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``pos`` is not implicitly convertible to a C double.

    :raises IndexError:
        If ``len(pos) < 3``.

    :raises TypeError:
        If ``pos`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc3`
    | :class:`forcedimension_core.containers.numpy.Enc3`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaEncoders()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncoderToPosition()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncodersToJointAngles()`
    """

    return _runtime._libdhd.dhdDeltaPositionToEncoder(
        pos[0], pos[1], pos[2], *out.ptrs, ID
    )


def deltaMotorToForce(
    mot: Array[int, int],
    enc: Array[int, int],
    out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the force applied to the end-effector
    about the X, Y, and Z axes (in [N]) for a given set of motor commands at a
    given position (defined by encoder readings)

    :param Array[int, int] mot:
        Sequence of ``(mot0, mot1, mot2)`` where ``mot0``, ``mot1``,
        and ``mot2`` are the axis 0, 1, and 2 DELTA motor commands,
        respectively.

    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to encoder values on axis 0, 1, and 2,
        respectively.

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the force applied to the end effector about
        the X, Y, and Z axes (in [N])

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises TypeError:
        If ``out`` does not support item assignment either
        because it's immutable or not subscriptable.

    :raises IndexError:
        If ``len(out) < 3``.

    :raises ctypes.ArgumentError:
        If any element of ``output`` is not implicitly convertible to a C
        ushort.

    :raises IndexError:
        If ``len(output) < 3``.

    :raises TypeError:
        If ``output`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C char.

    :raises IndexError:
        If ``len(enc) < 3``

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.deltaForceToMotor()`
    """

    return _runtime._libdhd.dhdDeltaMotorToForce(
        mot[0], mot[1], mot[2],
        enc[0], enc[1], enc[2],
        *out.ptrs,
        ID
    )


def deltaForceToMotor(
    f: Array[int, float],
    enc: Array[int, int],
    out: SupportsPtrs3[c_ushort],
    ID=-1
) -> int:
    """
    This routine computes and returns the motor commands necessary to obtain a
    given force on the end-effector at a given position (defined by encoder
    readings).

    :param Array[int, float] f:
        Sequence of ``(fx, fy, fz)`` where ``fx``, ``fy``, and ``fz`` are the
        force on the DELTA end-effector on the X, Y, and Z axes, respectively
        (in [N]).

    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to encoder values on axis 0, 1, and 2, respectively.

    :param SupportsPtrs3[ctypes.c_ushort] out:
        An output buffer to store the motor commands on axes 0, 1, and 2.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If any element of ``f`` is not implicitly convertible to a C double.

    :raises IndexError:
        If ``len(f) < 3``.

    :raises TypeError:
        If ``f`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_ushort]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.MOTOR_SATURATED`
        on success, and -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mot3`
    | :class:`forcedimension_core.containers.numpy.Mot3`
    | :func:`forcedimension_core.dhd.direct_expert.deltaMotorToForce()`
    """

    return _runtime._libdhd.dhdDeltaForceToMotor(
        f[0], f[1], f[2],
        enc[0], enc[1], enc[2],
        *out.ptrs,
        ID
    )


def wristEncoderToOrientation(
    enc: Array[int, int],
    out: SupportsPtrs3[c_double],
    ID: int = -1,
) -> int:
    """
    For devices with a wrist structure, compute the individual angle of each
    joint (in [rad]), starting with the one located nearest to the wrist base
    plate. For the :data:`forcedimension_core.constants.DeviceType.OMEGA6`
    and the :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT`
    devices, angles are computed with respect to their internal reference
    frame, which is rotated 45 degrees or π/4 radians about the Y axis. Please
    refer to your device user manual for more information on your device
    coordinate system.

    Note
    ----
    This feature only applies to devices with a wrist. See
    the :ref:`device_types` section for more details.


    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to wrist encoder values on the first, second, and
        third joint, respectively

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the wrist end-effector orientation about the
        first, second, and third wrist joints (in [rad]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible
        to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.wristOrientationToEncoder()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToEncoders()`
    """
    return _runtime._libdhd.dhdWristEncoderToOrientation(
        enc[0], enc[1], enc[2],
        *out.ptrs,
        ID
    )


def wristOrientationToEncoder(
    orientation: Array[int, float],
    out: SupportsPtrs3[c_int],
    ID: int = -1,
) -> int:
    """
    For devices with a wrist structure, compute the encoder values from the
    individual angle of each joint, starting with the one located nearest to
    the wrist plate base.

    Note
    ----
    For the :data:`forcedimension_core.constants.DeviceType.OMEGA6` and
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT` devices,
    angles must be expressed with respect to their internal reference frame,
    which is rotated 45 degrees or π/4 radians about the Y axis. Please refer
    to your device user manual for more information on your device coordinate
    system.


    Note
    ----
    This feature only applies to devices with an wrist. See
    the :ref:`device_types` section for more details.


    :param Array[int, float] orientation:
        Sequence of ``(oa, ob, og)`` where ``oa``, ``ob``, and ``og`` refer to
        wrist end effector orientation (in [rad]) around the X, Y, and Z axes,
        respectively.

    :param SupportsPtrs3[ctypes.c_int] out:
        An output buffer to store the encoder values.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).


    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_int]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``orientation`` is not implicitly convertible to a C
        double.

    :raises IndexError:
        If ``len(orientation) < 3``.

    :raises TypeError:
        If ``orientation`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc3`
    | :class:`forcedimension_core.containers.numpy.Enc3`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToOrientation()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToEncoders()`
    """

    return _runtime._libdhd.dhdWristOrientationToEncoder(
        orientation[0], orientation[1], orientation[2],
        *out.ptrs,
        ID
    )


def wristMotorToTorque(
    output: Array[int, int],
    enc: Array[int, int],
    out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the torque applied to the wrist
    end-effector about the, X, Y, and Z axes (in [Nm]) for a given set of motor
    commands at a given orientation (defined by encoder values)

    Note
    ----
    This feature only applies to devices with an active wrist. See
    the :ref:`device_types` section for more details.


    :param Array[int, int] cmd:
        Sequence of ``(cmd0, cmd1, cmd2)`` where ``cmd0``, ``cmd1``,
        and ``cmd2`` are the axis 0, 1, and 2 DELTA motor commands,
        respectively.

    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to encoder values on axis 0, 1, and 2, respectively.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the torques applied to the wrist about the
        X, Y, and Z axes (in [Nm]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``cmd`` is not implicitly convertible to a C ushort.

    :raises IndexError:
        If ``len(cmd) < 3``.

    :raises TypeError:
        If ``cmd`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1, otherwise.

    See Also
    --------
     | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.wristTorqueToMotor()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointTorquesExtrema()`
    """

    return _runtime._libdhd.dhdWristMotorToTorque(
        output[0], output[1], output[2],
        enc[0], enc[1], enc[2],
        *out.ptrs,
        ID
    )


def wristTorqueToMotor(
    t: Array[int, float],
    enc: Array[int, int],
    out: SupportsPtrs3[c_ushort],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the wrist motor commands necessary to
    obtain a given torque on the wrist end-effector about the X, Y, and Z axes
    (in [Nm]) at a given orientation (defined by encoder values)

    Note
    ----
    This feature only applies to the following devices:

    :data:`forcedimension_core.constants.DeviceType.SIGMA7_RIGHT`
    :data:`forcedimension_core.constants.DeviceType.SIGMA7_LEFT`


    :param Array[int, float] t:
        Sequence of ``(t0, t1, t2)`` where ``t0``, ``t1``, and ``t2`` are the
        DELTA axis torque commands (in [Nm]) for axes 0, 1, and 2,
        respectively.

    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to encoder values on axis 0, 1, and 2, respectively.

    :param SupportsPtrs3[ctypes.c_ushort] out:
        An output buffer to store the wrist motor commands.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_ushort]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``t`` is not implicitly convertible to a C double.

    :raises IndexError:
        If ``len(t) < 3``.

    :raises TypeError:
        If ``t`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise

    See Also
    --------
    | :class:`forcedimension_core.containers.Mot3`
    | :class:`forcedimension_core.containers.numpy.Mot3`
    | :func:`forcedimension_core.dhd.direct_expert.wristTorqueToMotor()`
    """

    return _runtime._libdhd.dhdWristTorqueToMotor(
        t[0], t[1], t[2],
        enc[0], enc[1], enc[2],
        *out.ptrs,
        ID
    )


def gripperMotorToForce(
    cmd: int,
    enc_wrist_grip: SupportsPtr[c_int],
    out: c_double,
    ID: int = -1
) -> int:
    """
    Computes the force applied to the end-effector (in [N]) for a given
    motor command.

    Note
    ----
    This feature only applies to devices with an active gripper. See
    the :ref:`device_types` section for more details.


    :param int output:
        Motor command on gripper axis.

    :param SupportsPtr[c_int] enc_wrist_grip:
        Buffer which contains the encoder values about wrist joints 0, 1, and
        2 as well as the encoder value of the gripper.

    :param c_ushort out:
        Output buffer to store the force applied to the end-effector (in [N]).

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If ``cmd`` is not implicitly convertible to a C ushort.

    :raises AttributeError:
        If ``enc_wrist_grip.ptr`` is not a valid attribute of
        ``enc_wrist_grip``

    :raises ctypes.ArgumentError:
        If ``enc_wrist_grip.ptr`` is not of type Pointer[c_int]

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc4`
    | :class:`forcedimension_core.containers.numpy.Enc4`
    | :func:`forcedimension_core.dhd.direct_expert.gripperForceToMotor()`
    """

    return _runtime._libdhd.dhdGripperMotorToForce(
        cmd, out, enc_wrist_grip.ptr, ID
    )


def setMot(cmds: SupportsPtr[c_ushort], mask: int = 0xff, ID: int = -1) -> int:
    """
    Program motor commands to a selection of motor channels.
    Particularly useful when using the generic controller
    directly, without a device model attached.

    :param SupportsPtr[c_ushort] cmds:
        An input buffer to store the motor command values.

    :param int mask:
        Bitwise mask of which motor should be set.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If ``cmds.ptr`` is not  of type ``Pointer[c_int]``

    :raises IndexError:
        If ``len(cmds) < MAX_DOF``.

    :raises ctypes.ArgumentError:
        If ``mask`` is not implicitly convertible to a C uchar.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns: 0 on success, -1 otherwise

    See Also
    --------
    | :class:`forcedimension_core.containers.DOFMot`
    | :class:`forcedimension_core.containers.numpy.DOFMot`
    | :data:`forcedimension_core.dhd.expert.setMotor()`
    | :data:`forcedimension_core.dhd.expert.setDeltaMotor()`
    | :data:`forcedimension_core.dhd.expert.setWristMotor()`
    | :data:`forcedimension_core.dhd.expert.setGripperMotor()`

    """

    return _runtime._libdhd.dhdSetMot(cmds.ptr, mask, ID)


def setJointTorques(q: SupportsPtr[c_double], mask: int = 0xff, ID: int = -1):
    """
    Sets all joint torques on all active axes.

    :param SupportsPtr[c_double] q:
        An input buffer to store the Joint torques for each
        degree-of-freedom (in [Nm]).

    :param int mask:
        Bitwise mask of which joints should be set

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If any element of ``cmds`` is not implicitly convertible
        to a C double.

    :raises IndexError:
        If ``len(cmds) < MAX_DOF``.

    :raises ctypes.ArgumentError:
        If ``q.ptr`` is not  of type ``Pointer[c_double]``

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.DOFFloat`
    | :class:`forcedimension_core.containers.numpy.DOFFloat`
    | :func:`forcedimension_core.dhd.expert.setDeltaJointTorques()`
    | :func:`forcedimension_core.dhd.expert.setWristJointTorques()`
    | :func:`forcedimension_core.dhd.expert.setForceAndWristJointTorques()`
    | :func:`forcedimension_core.dhd.expert.setForceAndWristJointTorquesAndGripperForce()`
    """

    return _runtime._libdhd.dhdSetJointTorques(q.ptr, mask, ID)


def preloadMot(
    cmds: SupportsPtr[c_ushort], mask: int = 0xff, ID: int = -1
) -> int:
    """
    Program motor commands to a selection of motor channels.
    Unlike :func:`forcedimension_core.dhd.expert.setMot()`, this
    function saves the requested commands internally for later
    application by calling
    :func:`forcedimension_core.dhd.setForce()` and the friends.

    :param SupportsPtr[c_ushort] cmds:
        An input buffer to store the motor command values.

    :param int mask:
        Bitwise mask of which motor should be set.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If ``cmds.ptr`` is not  of type ``Pointer[c_ushort]``

    :raises IndexError:
        If ``len(cmds) < MAX_DOF``.

    :raises TypeError:
        If ``cmds`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``mask`` is not implicitly convertible to a C char.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns: 0 on success, -1 otherwise

    See Also
    --------
    | :class:`forcedimension_core.containers.DOFMot`
    | :class:`forcedimension_core.containers.numpy.DOFMot`
    | :func:`forcedimension_core.dhd.direct_expert.setMot()`
    """

    return _runtime._libdhd.dhdPreloadMot(cmds.ptr, mask, ID)


def gripperForceToMotor(
    f: float,
    enc_wrist_grip: SupportsPtr[c_int],
    out: c_ushort,
    ID: int = -1
) -> Tuple[int, int]:
    """
    Given a desired force to be displayed by the force gripper, this routine
    computes and returns the refering motor command.

    Note
    ----
    This feature only applies to devices with an active gripper. See
    the :ref:`device_types` section for more details.


    :param int f:
        Force on the gripper end-effector (in [N]).

    :param Array[int, int] enc_wrist:
        An output buffer to store the wrist encoding readings.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If ``f`` is not implicitly convertible to a C double.

    :raises AttributeError:
        If ``enc_wrist_grip.ptr`` is not a valid attribute of
        ``enc_wrist_grip``

    :raises ctypes.ArgumentError:
        If ``enc_wrist_grip.ptr`` is not of type Pointer[c_int]

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        A tuple in the form ``(cmd, err)``.  ``cmd`` is the motor command on
        the gripper axis. ``err`` is 0 or
        :data:`forcedimension_core.constants.MOTOR_SATURATED` on success,
        -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc4`
    | :class:`forcedimension_core.containers.numpy.Enc4`
    | :func:`forcedimension_core.dhd.direct_expert.gripperForceToMotor()`
    """

    return _runtime._libdhd.dhdGripperForceToMotor(f, out, enc_wrist_grip.ptr, ID)


def getEnc(out: SupportsPtr[c_int], mask: int = 0xff, ID: int = -1) -> int:
    """
    Get a selective list of encoder values. Particularly useful when using the
    generic controller directly, without a device model attached.

    :param SupportsPtr[c_int] out:
        An output buffer to store the encoder values.

    :param int mask:
        Bitwise mask of which motor should be set.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not  of type ``Pointer[c_int]``

    :raises ctypes.ArgumentError:
        If ``mask`` is not implicitly convertible to a C uchar.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.DOFInt`
    | :class:`forcedimension_core.containers.numpy.DOFInt`
    | :func:`forcedimension_core.dhd.direct_expert.getEncVelocities()`
    """

    return _runtime._libdhd.dhdGetEnc(out.ptr, mask, ID)


def getDeltaJointAngles(out: SupportsPtrs3[c_double], ID: int = -1) -> int:
    """
    Retrieve the joint angles (in [rad]) for the DELTA structure.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the joint angles (in [rad]) of the DELTA
        structure for axes 0, 1, and 2.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToEncoders()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToJacobian()`
    """

    return _runtime._libdhd.dhdGetDeltaJointAngles(*out.ptrs, ID)


def getDeltaJacobian(out: SupportsPtr[c_double], ID: int = -1) -> int:
    """
    Retrieve the 3x3 jacobian matrix for the DELTA structure based on the
    current end-effector position. Please refer to your device user manual for
    more information on your device coordinate system.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtr[c_double] out:
        An output buffer to store the Jacobian.

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not of type Pointer[c_double]

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mat3x3`
    | :class:`forcedimension_core.containers.numpy.Mat3x3`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToJacobian()`
    """

    return _runtime._libdhd.dhdGetDeltaJacobian(out.ptr, ID)


def deltaJointAnglesToJacobian(
    joint_angles: Array[int, float],
    out: SupportsPtr[c_double],
    ID: int = -1,
) -> int:
    """
    Retrieve the 3x3 jacobian matrix for the DELTA structure
    based on a given joint configuration. Please refer to your device user
    manual for more information on your device coordinate system.

    :param Array[int, float] joint_angles:
        Sequence of (j0, j1, j2) where ``j0``, ``j1``, and ``j2`` refer to the
        joint angles (in [rad]) for axis 0, 1, and 2, respectively.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtr[c_double] out:
        An output buffer to store the Jacobian.

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not of type Pointer[c_double]

    :raises ctypes.ArgumentError:
        If any element of ``joint_angles`` is not implicitly convertible to C
        double.

    :raises IndexError:
        If ``len(joint_angles) < 3``.

    :raises TypeError:
        If ``joint_angles`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mat3x3`
    | :class:`forcedimension_core.containers.numpy.Mat3x3`
    | :func:`forcedimension_core.dhd.direct_expert.getJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaJacobian()`
    """

    return _runtime._libdhd.dhdDeltaJointAnglesToJacobian(
        joint_angles[0], joint_angles[1], joint_angles[2],
        out.ptr,
        ID
    )


def deltaJointTorquesExtrema(
    joint_angles: Array[int, float],
    minq_out: SupportsPtr[c_double],
    maxq_out: SupportsPtr[c_double],
    ID: int = -1
) -> int:
    """
    Compute the range of applicable DELTA joint torques (in [Nm]) for a
    given DELTA joint angle configuration (in [rad]).
    Please refer to your device user manual for more information on your device
    coordinate system.

    :param Array[int, float] joint_angles:
        Sequence of (j0, j1, j2) where ``j0``, ``j1``, and ``j2`` refer to the
        joint angles (in [rad]) for axis 0, 1, and 2, respectively.

    :param SupportsPtrs3[ctypes.c_double] minq_out:
        An output buffer to store the minimum appliable DELTA joint torques
        about axes 0, 1, and 2 (in [Nm]).

    :param SupportsPtrs3[ctypes.c_double] maxq_out:
        An output buffer to store the maximum appliable DELTA joint torques
        about axes 0, 1, and 2 (in [Nm]).

    :raises AttributeError:
        If ``minq_out.ptrs`` is not a valid attribute of ``minq_out``

    :raises TypeError:
        If ``minq_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``minq_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``maxq_out.ptrs`` is not a valid attribute of ``maxq_out``

    :raises TypeError:
        If ``maxq_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``maxq_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``joint_angles`` is not implicitly convertible to
        a C double.

    :raises IndexError:
        If ``len(joint_angles) < 3``.

    :raises TypeError:
        If ``joint_angles`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointTorquesExtrema()`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getJointAngles()`
    """

    return _runtime._libdhd.dhdDeltaJointTorquesExtrema(
        joint_angles[0],
        joint_angles[1],
        joint_angles[2],
        minq_out.ptr,
        maxq_out.ptr,
        ID
    )


def deltaEncodersToJointAngles(
    enc: Array[int, int],
    out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the DELTA joint angles (in [rad]) for a
    given set of encoder values.

    :param Array[int, int] enc:
        Sequence of ``(enc0, enc1, enc2)`` where ``enc0``, ``enc1``, and
        ``enc2`` refer to encoder values on axis 0, 1, and 2, respectively.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the DELTA joint angles (in [rad]) about axes
        0, 1, and 2.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    "ap"
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToEncoders()`
    """

    return _runtime._libdhd.dhdDeltaEncodersToJointAngles(
        enc[0],
        enc[1],
        enc[2],
        *out.ptrs,
        ID
    )


def deltaJointAnglesToEncoders(
    joint_angles: Array[int, float],
    out: SupportsPtrs3[c_int],
    ID: int = -1,
) -> int:
    """
    This routine computes and returns the DELTA encoder values for a given
    set of joint angles (in [rad]).

    :param Array[int, float] enc:
        Sequence of ``(j0, j1, j1)`` where ``j0``, ``j1``, and ``j2`` refer to
        DELTA joint angles (in [rad]) for axes 0, 1, and 2, respectively.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the DELTA encoder values.

    :raises ctypes.ArgumentError:
        If any element of ``joint_angles`` is not implicitly convertible to
        a C double.

    :raises IndexError:
        If ``len(joint_angles) < 3``.

    :raises TypeError:
        If ``joint_angles`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc3`
    | :class:`forcedimension_core.containers.numpy.Enc3`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaEncodersToJointAngles()`
    """

    return _runtime._libdhd.dhdDeltaJointAnglesToEncoders(
        joint_angles[0],
        joint_angles[1],
        joint_angles[2],
        *out.ptrs,
        ID
    )


def getWristJointAngles(out: SupportsPtrs3[c_double], ID: int = -1) -> int:
    """
    Retrieve the joint angles (in [rad]) for the wrist structure.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the wrist joint angles.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.dhdWristJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.dhdWristJointTorquesExtrema()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToEncoders()`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToJointAngles()`
    """

    return _runtime._libdhd.dhdGetWristJointAngles(*out.ptrs, ID)


def getWristJacobian(out: SupportsPtr[c_double], ID: int = -1) -> int:
    """
    Retrieve the 3x3 jacobian matrix for the wrist structure based on the
    current end-effector position. Please refer to your device user manual for
    more information on your device coordinate system.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtr[c_double] out:
        An output buffer to store the Jacobian.

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not a ``Pointer[c_double]`` type.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        and -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mat3x3`
    | :class:`forcedimension_core.containers.numpy.Mat3x3`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToJacobian()`
    """

    return _runtime._libdhd.dhdGetWristJacobian(out.ptr, ID)


def wristJointAnglesToJacobian(
    joint_angles: Array[int, float],
    out: SupportsPtr[c_double],
    ID: int = -1,
) -> int:
    """
    Retrieve the 3x3 jacobian matrix for the wrist structure
    based on a given joint configuration. Please refer to your device user
    manual for more information on your device coordinate system.

    :param Array[int, float] joint_angles:
        Sequence of ``(j0, j1, j2)`` where ``j0``, ``j1``, and ``j2`` refer to
        the joint angles for wrist axis 0, 1, and 2, respectively.

    :param SupportsPtr[c_double] out:
        An output buffer to store the wrist jacobian.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not a ``Pointer[c_double]`` type.

    :raises ctypes.ArgumentError:
        If any element of ``joint_angles`` is not implicitly convertible to
        a C double.

    :raises IndexError:
        If ``len(joint_angles) < 3``.

    :raises TypeError:
        If ``joint_angles`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mat3x3`
    | :class:`forcedimension_core.containers.numpy.Mat3x3`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getWristJointAngles()`
    """

    return _runtime._libdhd.dhdWristJointAnglesToJacobian(
        joint_angles[0],
        joint_angles[1],
        joint_angles[2],
        out.ptr,
        ID
    )


def wristJointTorquesExtrema(
    joint_angles: Array[int, float],
    minq_out: SupportsPtr[c_double],
    maxq_out: SupportsPtr[c_double],
    ID: int = -1
) -> int:
    """
    Compute the range of applicable wrist joint torques for a given wrist joint
    angle configuration. Please refer to your device user manual for more
    information on your device coordinate system.

    :param Array[int, float] joint_angles:
        Sequence of ``(j0, j1, j2)`` where ``j0``, ``j1``, ``j2`` refer to the
        joint angles for wrist axes 0, 1, and 2, respectively.

    :param SupportsPtrs3[ctypes.c_double] minq_out:
        An output buffer to store the minimum appliable DELTA joint torques
        about axes 0, 1, and 2 (in [Nm]).

    :param SupportsPtrs3[ctypes.c_double] maxq_out:
        An output buffer to store the maximum appliable DELTA joint torques
        about axes 0, 1, and 2 (in [Nm]).

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises ctypes.ArgumentError:
        If any element of ``joint_angles`` is not implicitly convertible to C
        double.

    :raises IndexError:
        If ``len(joint_angles) < 3``.

    :raises TypeError:
        If ``joint_angles`` is not subscriptable.

    :raises AttributeError:
        If ``minq_out.ptrs`` is not a valid attribute of ``minq_out``

    :raises TypeError:
        If ``minq_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``minq_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``maxq_out.ptrs`` is not a valid attribute of ``maxq_out``

    :raises TypeError:
        If ``maxq_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``maxq_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.getJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getWristJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToEncoders()`
    """

    return _runtime._libdhd.dhdWristJointTorquesExtrema(
        joint_angles[0],
        joint_angles[1],
        joint_angles[2],
        minq_out.ptr,
        maxq_out.ptr,
        ID
    )


_runtime._libdhd.dhdWristEncodersToJointAngles.argtypes = [
    c_int,
    c_int,
    c_int,
    c_double_ptr,
    c_double_ptr,
    c_double_ptr,
    c_byte
]
_runtime._libdhd.dhdWristEncodersToJointAngles.restype = c_int


def wristEncodersToJointAngles(
    enc: Array[int, int],
    out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the wrist joint angles (in [rad])
    for a given set of encoder values.

    :param Array[int, int] enc:
        Sequence of (enc0, enc1, enc2) where ``enc0``, ``enc1``, and ``enc2``
        refer to encoder values on wrist axes 0, 1, and 2, respectively.

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the joint angles (in [rad]) about wrist
        joint axes 0, 1, and 2.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``enc`` is not implicitly convertible to a C int.

    :raises IndexError:
        If ``len(enc) < 3``.

    :raises TypeError:
        If ``enc`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointTorquesExtrema()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToEncoders()`
    """

    return _runtime._libdhd.dhdWristEncodersToJointAngles(
        enc[0], enc[1], enc[2], *out.ptrs, ID
    )


def wristJointAnglesToEncoders(
    joint_angles: Array[int, float],
    out: SupportsPtrs3[c_int],
    ID: int = -1
) -> int:
    """
    This routine computes and returns the wrist encoder values for a given
    set of wrist joint angles (in [rad]).

    :param Array[int, float] enc:
        Sequence of ``(j0, j1, j1)`` where ``j0``, ``j1``, and ``j2`` refer to
        wrist joint angles (in [rad]) for axes 0, 1, and 2, respectively.

    :param Array[int, int] out:
        An output buffer to store the wrist encoder values.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_int]`` types.

    :raises ctypes.ArgumentError:
        If any element of ``joint_angles`` is not implicitly convertible to
        a C double.

    :raises IndexError:
        If ``len(joint_angles) < 3``.

    :raises TypeError:
        If ``joint_angles`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Enc3`
    | :class:`forcedimension_core.containers.numpy.Enc3`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.wristEncodersToJointAngles()`
    """

    return _runtime._libdhd.dhdWristJointAnglesToEncoders(
        joint_angles[0],
        joint_angles[1],
        joint_angles[2],
        *out.ptrs,
        ID
    )


def getJointAngles(out: SupportsPtr[c_double], ID: int = -1) -> int:
    """
    Retrieve the joint angles (in [rad]) for all sensed degrees-of-freedom of
    the current device.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtr[c_double] out:
        An output buffer to store the joint angles (in [rad]) for all sensed
        degrees-of-freedom for the current device.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of
        :data:`forcedimension_core.constants.MAX_DOF`
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        -1 otherwise.

    See Also
    ---------
    | :class:`forcedimension_core.containers.DOFFloat`
    | :class:`forcedimension_core.containers.numpy.DOFFloat`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getWristJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToEncoders()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToEncoders()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointTorquesExtrema()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointTorquesExtrema()`
    | :func:`forcedimension_core.dhd.direct_expert.deltaJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.wristJointAnglesToJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.jointAnglesToIntertiaMatrix()`
    | :func:`forcedimension_core.dhd.direct_expert.jointAnglesToGravityJointTorques()`
    | :func:`forcedimension_core.dhd.direct_expert.getJointVelocities()`
    """

    return _runtime._libdhd.dhdGetJointAngles(out.ptr, ID)


def getJointVelocities(out: SupportsPtr[c_double], ID: int = -1) -> int:
    """
    Retrieve the joint angle velocities (in [rad/s]) for all sensed
    degrees-of-freedom of the current device.

    :param SupportsPtr[c_double] out:
        An output buffer to store the joint velocities (in [rad/s]) for all
        sensed degrees-of-freedom of the current device.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of
        :data:`forcedimension_core.constants.MAX_DOF`
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
        -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.DOFFloat`
    | :class:`forcedimension_core.containers.numpy.DOFFloat`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.getJointAngles()`
    | :func:`forcedimension_core.dhd.direct_expert.getDeltaJacobian()`
    | :func:`forcedimension_core.dhd.direct_expert.getWristJacobian()`
    """

    return _runtime._libdhd.dhdGetJointVelocities(out.ptr, ID)


def getEncVelocities(out: SupportsPtr[c_double], ID: int = -1) -> int:
    """
    Retrieve the encoder angle velocities (in [inc/s]) for all sensed
    degrees-of-freedom of the current device.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtr[c_double] out:
        An output buffer to store the encoder velocities (in [inc/s]) for all
        sensed degrees-of-freedom of the current device.

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of
        :data:`forcedimension_core.constants.MAX_DOF`
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
       0 or :data:`forcedimension_core.constants.TIMEGUARD` on success,
       -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.DOFFloat`
    | :class:`forcedimension_core.containers.numpy.DOFFloat`
    | :func:`forcedimension_core.dhd.direct_expert.getEnc()`
    """

    return _runtime._libdhd.dhdGetEncVelocities(out.ptr, ID)


def jointAnglesToIntertiaMatrix(
    joint_angles: SupportsPtr[c_double],
    out: SupportsPtr[c_double],
    ID: int = -1,
) -> int:
    """
    Retrieve the inertia matrix based on a given joint
    configuration. Please refer to your device user manual for more information
    on your device coordinate system.

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param SupportsPtr[c_double] out:
        An array of joint angles (in [rad]) containing joint angles for all
        degrees of freedom.

    :param SupportsPtr[c_double] out:
        An output buffer for the 6x6 inertia matrix.

    :raises AttributeError:
        If ``joint_angles.ptrs`` is not a valid attribute of ``joint_angles``

    :raises TypeError:
        If ``joint_angles.ptrs`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``joint_angles.ptr`` is not of type ``Pointer[c_double]``.

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not a ``Pointer[c_double]`` type.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mat6x6`
    | :class:`forcedimension_core.containers.numpy.Mat6x6`
    | :func:`forcedimension_core.dhd.direct_expert.getJointAngles()`
    """

    return _runtime._libdhd.dhdJointAnglesToInertiaMatrix(
        joint_angles.ptr, out.ptr, ID
    )


def jointAnglesToGravityJointTorques(
    joint_angles: SupportsPtr[c_double],
    out: SupportsPtr[c_double],
    mask: int = 0xff,
    ID: int = -1,
) -> int:
    """
    This function computes the joint torques (in [Nm]) required to provide
    gravity compensation on all actuated degrees-of-freedom of the current
    device for a given joint angles configuration (in [rad])

    :param int ID:
        Device ID (see :ref:`multiple_devices` section for details).

    :param int mask:
        Bitwise mask of which joint torques should be computed.

    :param SupportsPtr[c_double] joint_angles:
        Output buffer for the joint torques required to provide gravity
        compensation (in [Nm]).

    :param SupportsPtr[c_double] out:
        Output buffer for the joint torques required to provide gravity
        compensation (in [Nm]).

    :raises AttributeError:
        If ``joint_angles.ptrs`` is not a valid attribute of ``joint_angles``

    :raises TypeError:
        If ``joint_angles.ptrs`` is not subscriptable.

    :raises ctypes.ArgumentError:
        If ``joint_angles.ptr`` is not of type ``Pointer[c_double]``.

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not a ``Pointer[c_double]`` type.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to a C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct_expert.getJointAngles()`
    """

    return _runtime._libdhd.dhdJointAnglesToGravityJointTorques(
        joint_angles.ptr,
        out.ptr,
        mask,
        ID
    )



from ctypes import c_bool, c_byte, c_double
from forcedimension_core import runtime as _runtime


_runtime._libdhd.dhdKbHit.argtypes = []
_runtime._libdhd.dhdKbHit.restype = c_bool


def kbHit() -> bool:
    """
    Check keyboard for a key hit. This function is OS
    independent.

    :returns:
        `True` if a key on the keyboard was hit, and `False`
        otherwise.
    """
    return _runtime._libdhd.dhdKbHit()


_runtime._libdhd.dhdKbGet.argtypes = []
_runtime._libdhd.dhdKbGet.restype = c_byte


def kbGet() -> str:
    """
    Retrieve a character from the keyboard. This function is OS
    independent.

    :returns:
        The character hit on the keyboard.
    """
    return chr(_runtime._libdhd.dhdKbGet())


_runtime._libdhd.dhdGetTime.argtypes = []
_runtime._libdhd.dhdGetTime.restype = c_double


def getTime() -> float:
    """
    Returns the current value from the high-resolution system
    counter in [s]. The resolution of the system counter may be
    machine-dependent, as it is usually derived from one of the
    CPU clock signals. The time returned, however, is guarunteed
    to be monotonic.

    :returns:
        The current monotonic time in [s] from the
        high-resolution system counter.
    """
    return _runtime._libdhd.dhdGetTime()


_runtime._libdhd.dhdSleep.argtypes = [c_double]
_runtime._libdhd.dhdSleep.restype = None


def sleep(sec: float) -> None:
    """
    Sleep for a given period of time in [s]. This function is OS
    independent.
    """
    _runtime._libdhd.dhdSleep(sec)



import ctypes
import getpass
import glob
import os
import pathlib
import platform
import sys
import unittest.mock as __mock
from typing import Final, Iterable, Set

from forcedimension_core.containers import VersionTuple

VERSION_TARGET = VersionTuple(3, 16, 0, 0)

SUPPORTED_PLATFORMS: Final[Set[str]] = {'linux', 'win32', 'cygwin', 'darwin'}

_ctypes_impl = ctypes
_getpass_impl = getpass
_glob_impl = glob
_os_impl = os
_pathlib_impl = pathlib
_platform_impl = platform
_sys_impl = sys

_test_load = False


def _get_search_paths_win32(search_dirs: Iterable[str] = ()):
    search_dirs = list(search_dirs)

    if _platform_impl.architecture()[0] == '64bit':
        lib_name = "drd64"
    else:
        lib_name = "drd"

    if _os_impl.environ.get('FDSDK'):
        search_dirs.append(
            _os_impl.path.join(
                _os_impl.environ.get(
                    'FDSDK'), 'bin', f'{lib_name}.dll'  # type: ignore
            )
        )

    if _sys_impl.platform == 'cygwin':
        root = _os_impl.path.join(
            _os_impl.path.sep + 'cygdrive',
            _pathlib_impl.Path.home().drive[0].lower()
        )
    else:
        root = _pathlib_impl.Path.home().drive + _os_impl.path.sep

    app_local = _os_impl.path.join(
        root, 'Users', _getpass_impl.getuser(), 'AppData', 'Local'
    )

    lib_versions = _glob_impl.glob(
        f'{_os_impl.path.join(root, "Program Files", "Force Dimension")}'
        f'{_os_impl.path.sep}sdk-*'
    )

    lib_versions_local = _glob_impl.glob(
        f'{_os_impl.path.join(app_local, "Force Dimension")}'
        f'{_os_impl.path.sep}sdk-*'
    )

    lib_versions.sort()
    lib_versions_local.sort()

    if lib_versions:
        search_dirs.append(
            _os_impl.path.join(
                lib_versions_local[-1], "bin", f"{lib_name}.dll"
            )
        )

    if lib_versions_local:
        search_dirs.append(
            _os_impl.path.join(lib_versions[-1], "bin", f"{lib_name}.dll")
        )

    return search_dirs


def _get_search_paths_unix(search_dirs: Iterable[str] = ()):
    search_dirs = list(search_dirs)
    machine = _platform_impl.machine()

    if _sys_impl.platform == 'linux':
        lib_file_glob = "libdrd.so.*"
        lib_file = "libdrd.so"
        platform_name = 'lin'
        compiler = 'gcc'
    else:
        lib_file_glob = "libdrd.*dylib"
        lib_file = "libdrd.dylib"
        platform_name = 'mac'
        compiler = 'clang'

    if (libpath := _os_impl.environ.get('FDSDK')):
        lib_folder = _os_impl.path.realpath(
            _os_impl.path.join(
                libpath,
                "lib",
                "release",
                f"{platform_name}-{machine}-{compiler}",  # noqa
            )
        )
        # type: ignore
        if (glob_res := _glob_impl.glob(f"{lib_folder}/{lib_file_glob}")):  # noqa
            glob_res.sort()
            search_dirs.append(glob_res[-1])

    # Legacy support for the old environment variable

    if (libpath := _os_impl.environ.get('FORCEDIM_SDK')):
        lib_folder = _os_impl.path.realpath(
            _os_impl.path.join(
                libpath,
                "lib",
                "release",
                f"{platform_name}-{machine}-{compiler}",  # type: ignore
            )
        )

        if (glob_res := _glob_impl.glob(f"{lib_folder}/{lib_file_glob}")):  # noqa
            glob_res.sort()
            search_dirs.append(glob_res[-1])

    lib_loc_local_glob = os.path.join(
        _os_impl.path.expanduser('~'),
        '.local', 'lib', lib_file_glob  # type: ignore
    )

    lib_loc_glob = os.path.join(
        _os_impl.path.sep, 'usr', 'local', 'lib', lib_file_glob  # type: ignore
    )

    lib_loc_local = os.path.join(
        _os_impl.path.expanduser('~'),
        '.local', 'lib', lib_file  # type: ignore
    )

    lib_loc = os.path.join(
        _os_impl.path.sep, 'usr', 'local', 'lib', lib_file  # type: ignore
    )

    glob_res_local = _glob_impl.glob(lib_loc_local_glob)
    glob_res = _glob_impl.glob(lib_loc_glob)

    search_dirs.append(lib_loc_local)

    if glob_res_local:
        glob_res_local.sort()
        search_dirs.append(glob_res_local[-1])

    search_dirs.append(lib_loc)
    if glob_res:
        glob_res.sort()
        search_dirs.append(glob_res[-1])

    return search_dirs  # type: ignore


def _get_search_paths(
    search_dirs: Iterable[str] = (), silent: bool = False
):
    if _sys_impl.platform == 'win32' or _sys_impl.platform == 'cygwin':
        return _get_search_paths_win32(search_dirs)
    elif _sys_impl.platform == 'linux' or _sys_impl.platform == 'darwin':
        return _get_search_paths_unix(search_dirs)
    else:
        if not silent:
            _sys_impl.stderr.write("Unsupported platform.\n")

        return []


def _get_version(lib: _ctypes_impl.CDLL):
    major = _ctypes_impl.c_int()
    minor = _ctypes_impl.c_int()
    release = _ctypes_impl.c_int()
    revision = _ctypes_impl.c_int()

    lib.dhdGetSDKVersion(
        _ctypes_impl.byref(major),
        _ctypes_impl.byref(minor),
        _ctypes_impl.byref(release),
        _ctypes_impl.byref(revision)
    )

    return VersionTuple(
        major.value,
        minor.value,
        release.value,
        revision.value
    )


def _should_mock():
    sphinx_build = (
        _os_impl.environ.get('__fdsdkpy_sphinx_build__', 'False') == 'True'
    )

    unittest = (
        _os_impl.environ.get('__fdsdkpy_unittest__', 'False') == 'True'
    )

    return sphinx_build or unittest


def load(
    search_dirs=(),
    silent=False,
):

    if _os_impl.environ.get('__fdsdkpy_unittest_runtime__', 'False') == 'True':
        return None

    if _should_mock() and not _test_load:
        return __mock.Mock()

    if not (search_dirs := _get_search_paths(search_dirs, silent)):
        return None

    for lib_path in search_dirs:
        if not _os_impl.path.isfile(lib_path):
            continue

        # Make sure to add the directory to PATH for Windows DLL loading
        if _sys_impl.platform == "win32":

            path = _os_impl.environ.get('PATH')
            directory, _ = _os_impl.path.split(lib_path)

            if not path:
                _os_impl.environ['PATH'] = directory
            else:
                if directory not in path:
                    _os_impl.environ['PATH'] = f"{directory};{path}"

        try:
            lib = _ctypes_impl.CDLL(lib_path)
        except OSError:
            if not silent:
                _sys_impl.stderr.write(
                    "Library was found but could not be loaded. Do you "
                    "have missing dependencies?\n"
                    "Ensure you have libusb-1."
                )

            return None
        if (version := _get_version(lib)) < VERSION_TARGET:  # type: ignore
            if not silent:
                _sys_impl.stderr.write(
                    f"Invalid version. v{version} found "
                    f"but v{VERSION_TARGET} is required.\n"
                )

            return None

        return lib  # type: ignore
    if not silent:
        _sys_impl.stderr.write(
            "Could not find libdrd. Is it installed?\n"  # type: ignore
        )

    return None


if (_libdrd_load := load()) is None:
    raise ImportError(
        "There were problems loading libdrd. Check if you have it installed "
        "properly. For more information on installation consult: "
        ""
    )

_libdrd = _libdrd_load
_libdhd = _libdrd_load



from enum import IntEnum
from typing import Final, Literal, Tuple

MAX_DOF: Final[int] = 8   #: Maximum number of DOF a device can have
MAX_BUTTONS: Final[int] = 32  #: Maximum number of buttons a device can have

VELOCITY_WINDOWING: Final[int] = 0  #: Velocity Windowing Mode
DEFAULT_VELOCITY_WINDOW: Final[int] = 20   #: Default velocity window size

MAX_STATUS: Final[int] = 17  #: Maximum number of elements in a status tuple

#: Return value used when the TimeGuard feature prevented an unnecessary
#: communication with the device.
TIMEGUARD: Final[int] = 1

#: Convinence constant used to set device to default timeguard value
#: Not the actual default timeguard value.
DEFAULT_TIMEGUARD_US: Final[int] = -1

#: Return value used when at least one of the motors cannot deliver the
#: requested torque. Motor groups are scaled in order to preserve force and
#: torque direction over magnitude.
MOTOR_SATURATED: Final[int] = 2


class VelocityEstimatorMode(IntEnum):
    """
    Supported velocity estimator modes. Currently only WINDOWING mode is
    available.
    """
    #: The default velocity estimator mode. In this mode, the velocity is
    #: estimated by comparing the current position with the position a given
    #: time interval ago. This time interval (or "window") can be adjusted
    #: using dhdConfigLinearVelocity, and should be modified to best suit the
    #: dynamic behavior of the device for a given application. The windowing
    #: estimator mode is the least resource intensive.
    WINDOWING = 0
    """
    The default velocity estimator mode. In this mode, the velocity is
    estimated by comparing the current position with the position a given
    time interval ago.
    """


class DeviceType(IntEnum):
    """
    Supported device type IDs as an enumeration.
    """
    NONE = 0
    DELTA3 = 63
    OMEGA3 = 33
    OMEGA6_RIGHT = 34
    OMEGA6_LEFT = 36
    OMEGA7_RIGHT = 35
    OMEGA7_LEFT = 37
    CONTROLLER = 81
    CONTROLLER_HR = 82
    CUSTOM = 91
    SIGMA3 = 206
    SIGMA7_RIGHT = 104
    SIGMA7_LEFT = 105
    LAMBDA3 = 203
    LAMBDA7_RIGHT = 108
    LAMBDA7_LEFT = 109
    FALCON = 60


class ErrorNum(IntEnum):
    """
    Error numbers reported by DHD as an enumeration.
    """
    NO_ERROR = 0
    ERROR = 1
    COM = 2
    DHC_BUSY = 3
    NO_DRIVER_FOUND = 4
    NO_DEVICE_FOUND = 5
    NOT_AVAILABLE = 6
    TIMEOUT = 7
    GEOMETRY = 8
    EXPERT_MODE_DISABLED = 9
    NOT_IMPLEMENTED = 10
    OUT_OF_MEMORY = 11
    DEVICE_NOT_READY = 12
    FILE_NOT_FOUND = 13
    CONFIGURATION = 14
    INVALID_INDEX = 15
    DEPRECATED = 16
    NULL_ARGUMENT = 17
    REDUNDANT_FAIL = 18
    NOT_ENABLED = 19
    DEVICE_IN_USE = 20
    INVALID = 21
    NO_REGULATION = 22


#: Array index for motors/encoders of the WRIST structure.
#: i.e. DELTA motor/encoder 0 has index 0
DELTA_IDX: Final[Tuple[Literal[0], Literal[1], Literal[2]]] = (0, 1, 2)


#: Array index for motors/encoders of the WRIST structure.
#: i.e. WRIST motor/encoder 0 has index 3
WRIST_IDX: Final[Tuple[Literal[3], Literal[4], Literal[5]]] = (3, 4, 5)


class NovintButtonID(IntEnum):
    """
    Enumeration mapping button type to button ID for the Novint Falcon.
    """
    CENTER = 0
    LEFT = 1
    UP = 2
    RIGHT = 3


class StatusIndex(IntEnum):
    """
    Index of statuses in the DHD status vector as an enumeration.
    """
    POWER = 0
    CONNECTED = 1
    STARTED = 2
    RESET = 3
    IDLE = 4
    FORCE = 5
    BRAKE = 6
    TORQUE = 7
    WRIST_DETECTED = 8
    ERROR = 9
    GRAVITY = 10
    TIMEGUARD = 11
    WRIST_INIT = 12
    REDUNDANCY = 13
    FORCEOFFCAUSE = 14


class ForceOffCause(IntEnum):
    """
    The event that caused forces to be disabled on the device (the last time
    forces were turned off).

    Info
    ----
    Not all devices suppport all the force-disabling mechanisms listed above.
    """
    FORCES_NOT_OFF = 0
    BUTTON = 1
    VELOCITY = 2
    WATCHDOG = 3
    SOFTWARE = 4
    USBDISCN = 5
    DEADMAN = 6


class ComMode(IntEnum):
    """
    Communication mode with the device. USB operations using
    :data:`forcedimension.constants.ComMode.SYNC` and
    :data:`forcedimension.constants.ComMode.ASYNC`.
    Other operation modes are reported for virtual devices
    (:data:`forcedimension.constants.ComMode.VIRTUAL`) and devices that are
    connected over the network
    (:data:`forcedimension.constants.ComMode.NETWORK`).
    """
    #: The synchronous USB mode performs USB read and write operations in
    #: sequence, allowing for a theoretical haptic refresh rate of 4 kHz.
    #: Please note that Other factors also influence USB performance, including
    #: the choice of operating system, machine load and program optimisation.
    SYNC = 0

    #: The asynchronous USB mode is the default. The asynchronous USB mode
    #: allows the operating system to parallelise the read and write operations
    #: on the USB port. This parallel operation improves refresh rate stability
    #: by reducing communication jitter. Other factors also influence USB
    #: performance, including the choice of operating system, machine load and
    #: program optimisation.
    ASYNC = 1

    #: This mode is reported when connected to a virtual device.
    VIRTUAL = 3

    #: This mode is reported when connected to a haptic device using the Force
    #: Dimension network connection mode.
    NETWORK = 4


class ThreadPriority(IntEnum):
    DEFAULT = 0
    HIGH = 1
    LOW = 2


class Handedness(IntEnum):
    NONE = 0
    LEFT = 1
    RIGHT = 2



from ctypes import c_double

import forcedimension_core.runtime as _runtime
from forcedimension_core.typing import (
    SupportsPtr, SupportsPtrs3
)


def getPosition(out: SupportsPtrs3[c_double], ID: int = -1) -> int:
    """
    Retrieve the position of the end-effector about the X, Y, and
    Z axes. Please refer to your device user manual for more
    information on your device coordinate system.

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the position of the end-effector
        (in [m]).

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.dhd.TIMEGUARD` on
        success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationDeg()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationFrame()`
    """

    return _runtime._libdhd.dhdGetPosition(*out.ptrs, ID)


def getForce(out: SupportsPtrs3[c_double], ID: int = -1) -> int:
    """
    Retrieve the force vector applied to the end-effector (in [N])
    about the X, Y, and Z axes Please refer to your device user
    manual for more information on your device coordinate system.

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the applied forces on the
        end-effector (in [N]).

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out`` does not support item assignment either
        because its immutable or not subscriptable.

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.setForce()`
    | :func:`forcedimension_core.dhd.direct.setForceAndTorque()`
    | :func:`forcedimension_core.dhd.direct.setForceAndTorqueAndGripperForce()`
    | :func:`forcedimension_core.dhd.direct.getForceAndTorque()`
    | :func:`forcedimension_core.dhd.direct.getForceAndTorqueAndGripperForce()`
    """

    return _runtime._libdhd.dhdGetForce(*out.ptrs, ID)


def getOrientationRad(
    out: SupportsPtrs3[c_double], ID: int = -1
) -> int:
    """
    For devices with a wrist structure, retrieve individual angle
    (in [rad]) of each joint, starting with the one located
    nearest to the wrist base plate.

    Note
    ----
    This feature only applies to devices with a wrist. See the
    :ref:`device_types` section for more details.


    Note
    ----
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_RIGHT`
    and
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT`
    have angles that are instead computed with respect to their
    internal reference frame, which is rotated π/4 radians around
    the Y axis. Please refer to your device user manual for more
    information on your device coordinate system.


    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the joint angles (in [rad]).

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out`` does not support item assignment either
        because its immutable or not subscriptable.

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getOrientationDeg()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationDeg()`
    """

    return _runtime._libdhd.dhdGetOrientationRad(*out.ptrs, ID)


def getOrientationDeg(
    out: SupportsPtrs3[c_double], ID: int = -1
) -> int:
    """
    For devices with a wrist structure, retrieve individual angle
    (in [deg]) of each joint, starting with the one located
    nearest to the wrist base plate.

    Note
    ----
    This feature only applies to devices with a wrist. See
    the :ref:`device_types` section for more details.


    Note
    ----
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_RIGHT`
    and
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT`
    have angles that are instead computed with respect to their
    internal reference frame, which is rotated π/4 radians around
    the Y axis. Please refer to your device user manual for more
    information on your device coordinate system.


    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the joint angles (in [deg]).

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out`` does not support item assignment either
        because its immutable or not subscriptable.

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationDeg()`
    """

    return _runtime._libdhd.dhdGetOrientationDeg(*out.ptrs, ID)


def getPositionAndOrientationRad(
    p_out: SupportsPtrs3[c_double],
    o_out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    Retrieve the position (in [m]) and
    for devices with a wrist structure, retrieve individual angle
    of each joint (in [rad]), starting with the one located nearest
    to the wrist base plate.

    Note
    ----
    This feature only applies to devices with a wrist. See
    the :ref:`device_types` section for more details.


    Note
    ----
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_RIGHT`
    and
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT`
    have angles that are instead computed with respect to their
    internal reference frame, which is rotated π/4 radians around
    the Y axis. Please refer to your device user manual for more
    information on your device coordinate system.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtrs3[ctypes.c_double] p_out:
        An output buffer to store the position (in [m]).

    :param SupportsPtrs3[ctypes.c_double] o_out:
        An output buffer to store the joint angles (in [rad]).

    :raises AttributeError:
        If ``p_out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``p_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``p_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``o_out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``o_out`` does not support item assignment either
        because its immutable or not subscriptable.

    :raises TypeError:
        If ``o_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``o_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getOrientationDeeg()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationDeg()`
    """

    return _runtime._libdhd.dhdGetPositionAndOrientationRad(
        *p_out.ptrs, *o_out.ptrs, ID
    )


def getPositionAndOrientationDeg(
    p_out: SupportsPtrs3[c_double],
    o_out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    Retrieve the position (in [m]) and
    for devices with a wrist structure, retrieve individual angle
    of each joint (in [deg]), starting with the one located nearest
    to the wrist base plate.


    Note
    ----
    This feature only applies to devices with a wrist. See
    the :ref:`device_types` section for more details.


    Note
    ----
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_RIGHT`
    and
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT`
    have angles that are instead computed with respect to their
    internal reference frame, which is rotated π/4 radians around
    the Y axis. Please refer to your device user manual for more
    information on your device coordinate system.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtrs3[ctypes.c_double] p_out:
        An output buffer to store the position (in [m]).

    :param SupportsPtrs3[ctypes.c_double] o_out:
        An output buffer to store the joint angles (in [rad]).

    :raises AttributeError:
        If ``p_out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``p_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``p_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``o_out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``o_out`` does not support item assignment either
        because its immutable or not subscriptable.

    :raises TypeError:
        If ``o_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``o_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getOrientationDeg()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationRad()`
    """

    return _runtime._libdhd.dhdGetPositionAndOrientationDeg(
        *p_out.ptrs, *o_out.ptrs, ID
    )


def getPositionAndOrientationFrame(
    p_out: SupportsPtrs3[c_double],
    matrix_out: SupportsPtr[c_double],
    ID: int = -1
) -> int:
    """
    Retrieve the position (in [m]) and orientation matrix of the
    end-effector about the X, Y, and Z axes. Please refer to your
    device user manual for more information on your device
    coordinate system.

    Note
    ----
    This feature only applies to devices with a wrist. See the
    :ref:`device_types` section for more details.


    Note
    ----
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_RIGHT`
    and
    :data:`forcedimension_core.constants.DeviceType.OMEGA6_LEFT`
    have angles that are instead computed with respect to their
    internal reference frame, which is rotated π/4 radians around
    the Y axis. Please refer to your device user manual for more
    information on your device coordinate system.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :raises AttributeError:
        If ``p_out.ptrs`` is not a valid attribute of ``p_out``

    :raises TypeError:
        If ``p_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``p_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``matrix_out.ptr`` is not a valid attribute of
        ``matrix_out``.

    :raises ctypes.ArgumentError:
        If ``p_out.ptr`` is not a ``Pointer[c_double]`` type.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getOrientationDeg()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationRad()`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationDeg()`
    """

    return _runtime._libdhd.dhdGetPositionAndOrientationFrame(
        *p_out.ptrs, matrix_out.ptr, ID
    )


def getForceAndTorque(
    f_out: SupportsPtrs3[c_double],
    t_out: SupportsPtrs3[c_double],
    ID: int = -1
) -> int:
    """
    Retrieve the forces (in [N]) and torques (in [Nm]) applied to
    the device end-effector about the X, Y, and Z axes. Please
    refer to your device user manual for more information on your
    device coordinate system.

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtrs3[ctypes.c_double] f_out:
        An output buffer to store the applied forces on the
        end-effector (in [N]).

    :param SupportsPtrs3[ctypes.c_double] t_out:
        An output buffer to store the applied torques on the
        end-effector (in [Nm]).

    :raises AttributeError:
        If ``f_out.ptrs`` is not a valid attribute of ``f_out``

    :raises TypeError:
        If ``f_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``f_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``t_out.ptrs`` is not a valid attribute of ``t_out``

    :raises TypeError:
        If ``t_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``t_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0, on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.setForce()`
    | :func:`forcedimension_core.dhd.direct.setForceAndTorque()`
    | :func:`forcedimension_core.dhd.direct.setForceAndTorqueAndGripperForce()`
    | :func:`forcedimension_core.dhd.direct.getForce()`
    | :func:`forcedimension_core.dhd.direct.getForceAndTorque()`
    | :func:`forcedimension_core.dhd.direct.getForceAndTorqueAndGripperForce()`
    """

    return _runtime._libdhd.dhdGetForceAndTorque(*f_out.ptrs, *t_out.ptrs, ID)


def getOrientationFrame(out: SupportsPtr[c_double], ID: int = -1) -> int:
    """
    Retrieve the rotation matrix of the wrist structure. The
    identity matrix is returned for devices that do not support
    orientations.

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtr[ctypes.c_double] out:
        An output buffer to store the orientation frame.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :raises AttributeError:
        If ``out.ptr`` is not a valid attribute of ``p_out``

    :raises ctypes.ArgumentError:
        If ``out.ptr`` is not a ``Pointer[c_double]`` type.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Mat3x3`
    | :class:`forcedimension_core.containers.numpy.Mat3x3`
    | :func:`forcedimension_core.dhd.direct.getPositionAndOrientationFrame()`
    """

    return _runtime._libdhd.dhdGetOrientationFrame(out.ptr, ID)


def getGripperThumbPos(
    out: SupportsPtrs3[c_double], ID: int = -1
) -> int:
    """
    Read the position (in [m]) of thumb rest location about the
    X, Y, and Z axes of the force gripper structure if present.

    Note
    ----
    This feature only applies to devices with a gripper. See
    the :ref:`device_types` section for more details.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the grippper thumb position
        (in [m]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getGripperFingerPos()`
    """

    return _runtime._libdhd.dhdGetGripperThumbPos(*out.ptrs, ID)


def getGripperFingerPos(
    out: SupportsPtrs3[c_double], ID: int = -1
) -> int:
    """
    Read the position (in [m]) of forefinger rest location about
    the X, Y, and Z axes of the force gripper structure if present.

    Note
    ----
    This feature only applies to devices with a gripper. See
    the :ref:`device_types` section for more details.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtrs3[ctypes.c_double] out:
        An output buffer to store the gripper finger position
        (in [m]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 or :data:`forcedimension_core.constants.TIMEGUARD`
        on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getGripperFingerPos()`
    """

    return _runtime._libdhd.dhdGetGripperFingerPos(*out.ptrs, ID)


def getForceAndTorqueAndGripperForce(
    f_out: SupportsPtrs3[c_double],
    t_out: SupportsPtrs3[c_double],
    fg_out: c_double,
    ID: int = -1
) -> int:
    """
    Retrieve the forces (in [N]) and torques (in [Nm]) applied
    to the device end-effector as well as the gripper force (in
    [N]). Forces and torques are about the X, Y, and Z axes.

    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtrs3[ctypes.c_double] f_out:
        An output buffer to store the applied forces on the
        end-effector (in [N]).

    :param SupportsPtrs3[ctypes.c_double] t_out:
        An output buffer to store the applied torques on the
        end-effector (in [Nm]).

    :raises AttributeError:
        If ``f_out.ptrs`` is not a valid attribute of ``f_out``

    :raises TypeError:
        If ``f_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``f_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``t_out.ptrs`` is not a valid attribute of ``t_out``

    :raises TypeError:
        If ``t_out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``t_out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises AttributeError:
        If ``fg_out.ptr`` is not a valid attribute of ``fg_out``

    :raises ctypes.ArgumentError:
        If ``fg_out.ptr`` is not a ``Pointer[c_double]``.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
       0, on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.getForce()`
    | :func:`forcedimension_core.dhd.direct.getForceAndTorque()`
    """

    err = _runtime._libdhd.dhdGetForceAndTorqueAndGripperForce(
        *f_out.ptrs, *t_out.ptrs, fg_out, ID
    )
    return err


def getLinearVelocity(out: SupportsPtrs3[c_double], ID: int = -1) -> int:
    """
    Retrieve the estimated instanteous linear velocity (in [m/s]).

    Note
    ----
    The velocity estimator requires at least 2 position
    updates during the time interval defined in
    :func:`forcedimension_core.dhd.configLinearVelocity()` in
    order to be able to compute the estimate. Otherwise, if there
    are no calls to :func:`forcedimension_core.dhd.getPosition()`
    or :func:`forcedimension_core.dhd.getLinearVelocity()`, this
    function will error with
    :data:`forcedimension_core.constants.ErrorNum.TIMEOUT`.
    For more information refer to :ref:`velocity_estimator`.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param SupportsPtr[ctypes.c_double] out:
        An output buffer to store the linear velocity (in [m/s]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.configLinearVelocity()`
    """

    return _runtime._libdhd.dhdGetLinearVelocity(*out.ptrs, ID)


def getAngularVelocityRad(
    out: SupportsPtrs3[c_double], ID: int = -1
) -> int:
    """
    Retrieve the estimated angular velocity (in [rad/s]).

    Note
    ----
    The velocity estimator requires at least 2 position
    updates during the time interval defined in
    :func:`forcedimension_core.dhd.configAngularVelocity()` in
    order to be able to compute the estimate. Otherwise, if there
    are no
    calls to:func:`forcedimension_core.dhd.getPosition()` or
    :func:`forcedimension_core.dhd.getAngularVelocityRad()`,
    this function will return error with
    :data:`forcedimension_core.constants.ErrorNum.TIMEOUT`.
    For more information refer to :ref:`velocity_estimator`.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param out:
        An output buffer to store the angular velocity (in [rad/s]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.


    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.configAngularVelocity()`
    | :func:`forcedimension_core.dhd.direct.getAngularVelocityDeg()`
    """

    return _runtime._libdhd.dhdGetAngularVelocityRad(*out.ptrs, ID)


def getAngularVelocityDeg(
    out: SupportsPtrs3[c_double], ID: int = -1
) -> int:
    """
    Retrieve the estimated angular velocity (in [deg/s]).

    Note
    ----
    The velocity estimator requires at least 2 position
    updates during the time interval defined in
    :func:`forcedimension_core.dhd.configAngularVelocity()` in
    order to be able to compute the estimate. Otherwise, if there
    are no calls to:func:`forcedimension_core.dhd.getPosition()`
    or :func:`forcedimension_core.dhd.getAngularVelocityDeg()`,
    this function will error with
    :data:`forcedimension_core.constants.ErrorNum.TIMEOUT`.
    For more information refer to :ref:`velocity_estimator`.


    :param int ID:
         Device ID (see :ref:`multiple_devices` section for
         details).

    :param out:
        An output buffer to store the angular velocity (in [deg/s]).

    :raises AttributeError:
        If ``out.ptrs`` is not a valid attribute of ``out``

    :raises TypeError:
        If ``out.ptrs`` is not iterable.

    :raises ctypes.ArgumentError:
        If ``out.ptrs`` does not expand into a tuple of 3
        ``Pointer[c_double]`` types.

    :raises ctypes.ArgumentError:
        If ``ID`` is not implicitly convertible to C char.

    :returns:
        0 on success, -1 otherwise.

    See Also
    --------
    | :class:`forcedimension_core.containers.Vec3`
    | :class:`forcedimension_core.containers.numpy.Vec3`
    | :func:`forcedimension_core.dhd.direct.configAngularVelocity()`
    | :func:`forcedimension_core.dhd.direct.getAngularVelocityRad()`
    """

    return _runtime._libdhd.dhdGetAngularVelocityDeg(*out.ptrs, ID)



from __future__ import annotations

import ctypes as ct
from array import array
from ctypes import c_int
import os
from typing import Any, Iterable, NamedTuple, Tuple
from typing_extensions import overload

import pydantic as pyd
import pydantic_core as pyd_core
from pydantic_core import core_schema as _core_schema

from forcedimension_core.constants import MAX_DOF, MAX_STATUS
from forcedimension_core.typing import (
    CBoolLike, Pointer, c_double_ptr, c_int_ptr, c_ushort_ptr
)

try:
    if os.environ.get('__fdsdk__unittest_opt_has_numpy__', 'True') == 'False':
        raise ImportError

    import forcedimension_core.containers.numpy as numpy
except ImportError:
    pass


class VersionTuple(NamedTuple):
    """
    Adapts the four seperate number return into a single grouped
    :class:`typing.NamedTuple`.
    """

    major: int
    minor: int
    release: int
    revision: int

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.release}-{self.revision}"


class Status(ct.Structure):
    """
    Adapts the status array returned by
    :func:`forcedimension.bindings.dhd.getStatus()`
    """

    @overload
    def __init__(self, status: Status):
        ...

    @overload
    def __init__(
        self,
        power: CBoolLike = False, connected: CBoolLike = False,
        started: CBoolLike = False, reset: CBoolLike = False,
        idle: CBoolLike = False, force: CBoolLike = False,
        brake: CBoolLike = False, torque: CBoolLike = False,
        wrist_detected: CBoolLike = False,
        error: int = False, gravity: CBoolLike = False,
        timeguard: CBoolLike = False, wrist_init: CBoolLike = False,
        redundancy: CBoolLike = False, forceoffcause: int = 0,
        locks: CBoolLike = False, axis_checked: CBoolLike = False
    ):
        ...

    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            if isinstance(args[0], Status):
                super().__init__(*args[0])
            elif isinstance(args[0], int) or isinstance(args[0], bool):
                super().__init__(args[0])
            else:
                raise ValueError(
                    "Cannot copy construct from an object not of type Status."
                )
        else:
            super().__init__(*args, **kwargs)

        self._ptr = ct.cast(ct.pointer(self), c_int_ptr)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda status: {
                    field: getattr(status, field)
                    for field in map(
                        lambda fields: fields[0], status._fields_[:-1]
                    )
                }
            )
        )

    @property
    def ptr(self) -> Pointer[c_int]:
        return self._ptr

    def __len__(self) -> int:
        return MAX_STATUS

    def __getitem__(self, i: int) -> int:
        return getattr(self, self._fields_[i][0])

    def __setitem__(self, i: int, val: int):
        setattr(self, self._fields_[i][0], val)

    def __iter__(self):
        for field_name, _ in self._fields_[:-1]:
            yield getattr(self, field_name)

    def __str__(self) -> str:
        return (
            "Status(power={}, connected={}, started={}, reset={}, idle={}, "
            "force={}, brake={}, torque={}, wrist_detected={}, error={}, "
            "gravity={}, timeguard={}, wrist_init={}, redundancy={}, "
            "forceoffcause={}, locks={}, axis_checked={})".format(
                *self
            )
        )

    _fields_ = (
        ('power', c_int),
        ('connected', c_int),
        ('started', c_int),
        ('reset', c_int),
        ('idle', c_int),
        ('force', c_int),
        ('brake', c_int),
        ('torque', c_int),
        ('wrist_detected', c_int),
        ('error', c_int),
        ('gravity', c_int),
        ('timeguard', c_int),
        ('wrist_init', c_int),
        ('redundancy', c_int),
        ('forceoffcause', c_int),
        ('locks', c_int),
        ('axis_checked', c_int),

        # Prevents out of bounds access if they expand MAX_STATUS.
        ('__padding', c_int * (32 - MAX_STATUS)),
    )

    #: Indicates if the device is powered or not.
    power: int
    "Indicates if the device is powered or not."

    #: Indicates if the device is connected or not.
    connected: int
    "Indicates if the device is connected or not."

    #: Indicates if the device controller is running or not.
    started: int
    "Indicates if the device controller is running or not."

    #: Indicates if the device is in RESET mode or not.
    #: See device modes for details.
    reset: int
    "1 if the device controller is in RESET mode, 0 otherwise"

    #: Indicates if the device is in IDLE mode or not.
    #: see device modes for details.
    idle: int
    "1 if the device controller is in idle mode or not, 0 otherwise"

    #: Indicates if the device is in force mode or not.
    #: see device modes for details.
    force: int
    "1 if the device controller is in force mode or not, 0 otherwise"

    #: Indicates if the device is in brake mode or not.
    #: see device modes for details.
    brake: int
    "1 if device controller is in break mode or not, 0 otherwise"

    #: indicates if the torques are active or not when the device is
    #: in force mode. see device modes for details.
    torque: int
    "1 if torques are active when the device is in force mode, 0 otherwise"

    #: Indicates if the device has a wrist or not.
    #: see device types for details.
    wrist_detected: int
    "1 if the device has a wrist, 0 otherwise."

    #: Indicates if the an error happened on the device controller.
    error: int
    "1 if an error happend on the device controller, 0 otherwise"

    #: Indicates if the gravity compensation option is enabled or not.
    gravity: int
    "1 if the gravity compensation option is enabled, 0 otherwise"

    #: Indicates if the TimeGuard op is enabled or not.
    #: See TimeGuard op for details.
    timeguard: int
    "1 if the TimeGuard option is enabled, 0 otherwise"

    #: Indicates if the device wrist is initialized or not.
    #: See device types for details.
    wrist_init: int
    "1 if the device wrist is initialized, 0 otherwise."

    #: The status of the redundant encoder consistency check. For devices
    #: equipped with redundant encoders, a value of 1 indicates that the
    #: redundancy check is successful. A value of 0 is reported otherwise, or
    # if the device does not op redundant encoders.
    redundancy: int
    """
    1 if the redundant encoder check was successful. For devices that
    don't op redundant encoders, this value is 0.
    """

    #: The event that caused forces to be disabled on the device (the last time
    #: forces were turned off).
    forceoffcause: int
    """
    The event that caused forces to be disabled on the device (the last time
    forces were turned off).
    """

    #: The status of the locks on supported devices. The value can be either
    #: 1 if the locks are engaged, 0 if the locks are disengagned,
    #: or -1 if the status of the locks is unknown.
    locks: int
    """
    The status of the locks on supported devices. The value can be either
    1 if the locks are engaged, 0 if the locks are disengagned, or -1 if the
    status of the locks is unknown.
    """

    #: A bit vector that indicates the validation status of each axis. The
    #: validation status of all device axes can be assessed by calling the
    #: :func:`forcedimension.drd.checkInit()` function in the Force Dimension
    #: Robotic SDK (DRD). Each bit of the status value returned corresponds
    #: to the validation status of the corresponding axis.
    axis_checked: int
    """
    A bit vector that indicates the validation status of each axis. The
    validation status of all device axes can be assessed by calling the
    drd.checkInit() function in the Force Dimension Robotic SDK (DRD). Each bit
    of the status value returned corresponds to a the validation status of the
    corresponding axis:
    """


class Vec3(array):
    """
    Represents an array of three C floats as a
    :class:`array.array`. Typically used by functions which
    return a vector or take a vector. Can also be used by
    functions which convert a set of 3 float such as in the
    functions which get orientation. Has convinence "x", "y",
    and "z" properties to access the 0th, 1st, and 2nd elements,
    respectively.
    """

    def __new__(
        cls, initializer: Iterable[float] = (0., 0., 0.)
    ):
        arr = super(Vec3, cls).__new__(
            cls, 'd', initializer  # type: ignore
        )

        if len(arr) != 3:
            raise ValueError(
                "Vector 3 initializer must be only contain 3 "
                "elements"
            )

        return arr

    def __init__(self, *args, **kwargs):
        ptr = self.buffer_info()[0]
        self._ptrs = (
            ct.cast(ptr, c_double_ptr),
            ct.cast(ptr + self.itemsize, c_double_ptr),
            ct.cast(ptr + 2 * self.itemsize, c_double_ptr),
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_double_ptr:
        """
        A pointer to the front of the array.
        """

        return self._ptrs[0]

    @property
    def ptrs(self) -> Tuple[c_double_ptr, c_double_ptr, c_double_ptr]:
        """
        A tuple of pointers to each element of the array in order.
        """

        return self._ptrs

    @property
    def x(self) -> float:
        """
        Alias to 0th element
        """

        return self[0]

    @x.setter
    def x(self, value: float):
        self[0] = value

    @property
    def y(self) -> float:
        """
        Alias to 1st element
        """

        return self[1]

    @y.setter
    def y(self, value: float):
        self[1] = value

    @property
    def z(self) -> float:
        """
        Alias to 2nd element
        """

        return self[2]

    @z.setter
    def z(self, value: float):
        self[2] = value


class Enc3(array):
    """
    Represents an array of three C ints as a :class:`array.array`.
    Typically used by functions which return information about
    encoders from the WRIST or DELTA structure.
    """

    def __new__(
        cls, initializer: Iterable[int] = (0, 0, 0)
    ):
        arr = super(Enc3, cls).__new__(
            cls, 'i', initializer  # type: ignore
        )

        if len(arr) != 3:
            raise ValueError(
                "Vector 3 initializer must be only contain 3 "
                "elements"
            )

        return arr

    def __init__(self, *args, **kwargs):
        ptr = self.buffer_info()[0]
        self._ptrs = (
            ct.cast(ptr, c_int_ptr),
            ct.cast(ptr + self.itemsize, c_int_ptr),
            ct.cast(ptr + 2 * self.itemsize, c_int_ptr),
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_int_ptr:
        """
        A pointer to the front of the array.
        """

        return self._ptrs[0]

    @property
    def ptrs(self) -> Tuple[c_int_ptr, c_int_ptr, c_int_ptr]:
        """
        A tuple of pointers to each element of the array in order.
        """

        return self._ptrs


class Mot3(array):
    """
    Represents an array of three C ushorts as a
    :class:`array.array`. Typically used functions which take
    motor commands or convert motor commands to forces
    (and vice versa). In those functions, represents an array
    of motor commands for each axis of the delta or wrist
    structure.
    """

    def __new__(
        cls, initializer: Iterable[int] = tuple(0 for _ in range(3))
    ):
        arr = super(Mot3, cls).__new__(
            cls, 'H', initializer  # type: ignore
        )

        if len(arr) != 3:
            raise ValueError()

        return arr

    def __init__(self, *args, **kwargs):
        ptr = self.buffer_info()[0]
        self._ptrs = (
            ct.cast(ptr, c_ushort_ptr),
            ct.cast(ptr + self.itemsize, c_ushort_ptr),
            ct.cast(ptr + 2 * self.itemsize, c_ushort_ptr),
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_ushort_ptr:
        """
        A pointer to the front of the array.
        """

        return self._ptrs[0]

    @property
    def ptrs(self) -> Tuple[c_ushort_ptr, c_ushort_ptr, c_ushort_ptr]:
        """
        A tuple of pointers to each element of the array in order.
        """

        return self._ptrs


class Enc4(array):
    """
    Represents an array of four C ints as a
    :class:`array.array`. Typically used functions which convert
    gripper motor commands to forces and vice versa.
    """

    def __new__(
        cls, initializer: Iterable[int] = (0, 0, 0, 0)
    ):
        arr = super(Enc4, cls).__new__(
            cls, 'i', initializer  # type: ignore
        )

        if len(arr) != 4:
            raise ValueError()

        return arr

    def __init__(self, *args, **kwargs):
        self._ptr = ct.cast(self.buffer_info()[0], c_int_ptr)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_int_ptr:
        """
        A pointer to the front of the underlying contiguous data.
        """

        return self._ptr


class DOFInt(array):
    """
    Represents an array of C ints, one for each
    degree-of-freedom  as a Python :class:`array.array`.
    Typically used by functions that get encoder values for each
    degree-of-freedom.
    """

    def __new__(
        cls, initializer: Iterable[int] = tuple(0 for _ in range(MAX_DOF))
    ):
        arr = super(DOFInt, cls).__new__(
            cls, 'i', initializer  # type: ignore
        )

        if len(arr) != MAX_DOF:
            raise ValueError()

        return arr

    def __init__(self, *args, **kwargs):
        self._ptr = ct.cast(self.buffer_info()[0], c_int_ptr)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_int_ptr:
        """
        A pointer to the front of the underlying contiguous data.
        """

        return self._ptr


class DOFMotor(array):
    """
    Represents an array of C unsigned shorts, one for each
    degree-of-freedom  as a Python :class:`array.array`.
    Typically used by functions that request motor commands for
    each degree-of-freedom.
    """

    def __new__(
        cls, initializer: Iterable[int] = tuple(0 for _ in range(MAX_DOF))
    ):
        arr = super(DOFMotor, cls).__new__(
            cls, 'H', initializer  # type: ignore
        )

        if len(arr) != MAX_DOF:
            raise ValueError()

        return arr

    def __init__(self, *args, **kwargs):
        self._ptr = ct.cast(self.buffer_info()[0], c_ushort_ptr)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_ushort_ptr:
        """
        A pointer to the front of the underlying contiguous data.
        """

        return self._ptr


class DOFFloat(array):
    """
    Represents an array of floats, one for each
    degree-of-freedom as a :class:`array.array`. Typically
    used by functions that request joint angles or linear/angular
    velocities for each  degree-of-freedom.
    """

    def __new__(
        cls, initializer: Iterable[float] = tuple(0 for _ in range(MAX_DOF))
    ):
        arr = super(DOFFloat, cls).__new__(
            cls, 'd', initializer  # type: ignore
        )

        if len(arr) != MAX_DOF:
            raise ValueError()

        return arr

    def __init__(self, *args, **kwargs):
        self._ptr = ct.cast(self.buffer_info()[0], c_double_ptr)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: arr.tolist()
            )
        )

    @property
    def ptr(self) -> c_double_ptr:
        """
        A pointer to the front of the underlying contiguous data.
        """

        return self._ptr


class Mat3x3(array):
    """
    Represents the type of a 3x3 matrix of floats
    :class:`array.array`. Typically used to represent a 3x3
    coordinate frame matrix.
    """

    def __new__(
        cls, initializer: Iterable[float] = tuple(0. for _ in range(9))
    ):
        arr = super(Mat3x3, cls).__new__(
            cls, 'd', initializer  # type: ignore
        )

        if len(arr) != 9:
            raise ValueError(
                "Mat3x3 initializer must be only contain 9 "
                "elements"
            )

        return arr

    def __init__(self, *args, **kwargs):
        self._ptr = ct.cast(self.buffer_info()[0], c_double_ptr)

    def __getitem__(self, indicies: Tuple[int, int]) -> float:
        if not isinstance(indicies, Tuple):
            raise TypeError("Indicies must be a tuple of two ints")

        i, j = indicies

        if not isinstance(i, int):
            raise TypeError("First index is not an int.")

        if not isinstance(j, int):
            raise TypeError("Second index is not an int.")

        return super().__getitem__(3 * i + j)

    def __setitem__(self, indicies: Tuple[int, int], value: float):
        if not isinstance(indicies, Tuple):
            raise TypeError("Indicies must be a tuple of two ints")

        i, j = indicies

        if not isinstance(i, int):
            raise TypeError("First index is not an int.")

        if not isinstance(j, int):
            raise TypeError("Second index is not an int.")

        super().__setitem__(3 * i + j, value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: [[arr[i, j] for j in range(3)] for i in range(3)]
            )
        )

    @property
    def ptr(self) -> c_double_ptr:
        """
        A pointer to the front of the underlying contiguous data.
        """

        return self._ptr


class Mat6x6(array):
    """
    Represents the type of a 6x6 matrix of floats
    :class:`array.array`. Typically used to represent a 6x6
    inertia matrix.
    """

    def __new__(
        cls, initializer: Iterable[float] = tuple(0. for _ in range(36))
    ):
        arr = super(Mat6x6, cls).__new__(
            cls, 'd', initializer  # type: ignore
        )

        if len(arr) != 36:
            raise ValueError(
                "Mat3x3 initializer must be only contain 36 "
                "elements"
            )

        return arr

    def __init__(self, *args, **kwargs):
        self._ptr = ct.cast(self.buffer_info()[0], c_double_ptr)

    def __getitem__(self, indicies: Tuple[int, int]) -> float:
        if not isinstance(indicies, Tuple):
            raise TypeError("Indicies must be a tuple of two ints")

        i, j = indicies

        if not isinstance(i, int):
            raise TypeError("First index is not an int.")

        if not isinstance(j, int):
            raise TypeError("Second index is not an int.")

        return super().__getitem__(6 * i + j)

    def __setitem__(self, indicies: Tuple[int, int], value: float):
        if not isinstance(indicies, Tuple):
            raise TypeError("Indicies must be a tuple of two ints")

        i, j = indicies

        if not isinstance(i, int):
            raise TypeError("First index is not an int.")

        if not isinstance(j, int):
            raise TypeError("Second index is not an int.")

        super().__setitem__(6 * i + j, value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pyd.GetCoreSchemaHandler
    ) -> pyd_core.CoreSchema:
        return _core_schema.no_info_plain_validator_function(
            cls,
            serialization=_core_schema.plain_serializer_function_ser_schema(
                lambda arr: [[arr[i, j] for j in range(6)] for i in range(6)]
            )
        )

    @property
    def ptr(self) -> c_double_ptr:
        """
        A pointer to the front of the underlying contiguous data.
        """

        return self._ptr



from enum import IntEnum


class DeviceType(IntEnum):
    OMEGA33 = 34
    OMEGA33_LEFT = 36
    OMEGA331 = 35
    OMEGA331_LEFT = 37
    SIGMA331 = 104
    SIGMA331_LEFT = 105
    LAMBDA331 = 108
    LAMBDA331_LEFT = 109
